/***************************************************************************
 *   copyright       : (C) 2012-2021 by Pascal Brachet                     *
 *   https://www.xm1math.net/pdfadd/                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   @license GPL-2.0+ <https://spdx.org/licenses/GPL-2.0+.html>           *
 ***************************************************************************/

#include "pdfadd.h"

#include "courbesdialog.h"
#include "tabvardialog.h"
#include "arbredialog.h"
#include "geometriedialog.h"
#include "graphedialog.h"
#include "boitesdialog.h"
#include "histodialog.h"
#include "suitesdialog.h"
#include "espacedialog.h"
#include "binomdialog.h"
#include "trigodialog.h"

#include "listdelegate.h" 

#if defined(Q_OS_UNIX) && !defined(Q_OS_MAC)
#include "x11fontdialog.h"
#endif

#include <QSettings>
#include <QAction>
#include <QStatusBar>
#include <QMessageBox>
#include <QApplication>
#include <QDir>
#include <QStyleFactory>
#include <QStyle>
#include <QDomDocument>
#include <QDomElement>
#include <QFile>
#include <QFileDialog>
#include <QFileInfo>
#include <QFontDatabase>
#include <QDebug>
#include <QTextStream>
#include <QScreen>
#include <QWindow>
#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
    #include <QtCore/QTextCodec>
#else
    #include <QtCore5Compat/QTextCodec>
#endif
#include <QColorDialog>

#define STRINGIFY_INTERNAL(x) #x
#define STRINGIFY(x) STRINGIFY_INTERNAL(x)

#define VERSION_STR STRINGIFY(PDFADDVERSION)

#if defined(Q_OS_WIN)
    QString searchInWindowPath(const QString &filter = QString())
    {
    QByteArray pEnv = qgetenv("PATH");
    const QLatin1Char pathSep(';');
    QStringList searchPaths;
    const QStringList rawPaths = QString::fromLocal8Bit(pEnv.constData()).split(pathSep, QString::SkipEmptyParts);
    searchPaths.reserve(rawPaths.size());
    foreach (const QString &rawPath, rawPaths) {
        QString cleanPath = QDir::cleanPath(rawPath);
        if (cleanPath.size() > 1 && cleanPath.endsWith(QLatin1Char('/')))
            cleanPath.truncate(cleanPath.size() - 1);
        searchPaths.push_back(cleanPath);
    }
    QStringList path_found=searchPaths.filter(filter);
    if  (!path_found.empty()) return path_found.takeLast();
    else return QString();
    }

    enum REG_SEARCH_TYPE
    {
        USER,
        SYSTEM_32BIT,
        SYSTEM_64BIT
    };

    QStringList getRegSubkeys(const HKEY handle)
    {
        QStringList keys;

        DWORD cSubKeys = 0;
        DWORD cMaxSubKeyLen = 0;
        LONG res = ::RegQueryInfoKeyW(handle, NULL, NULL, NULL, &cSubKeys, &cMaxSubKeyLen, NULL, NULL, NULL, NULL, NULL, NULL);

        if (res == ERROR_SUCCESS) {
            ++cMaxSubKeyLen; // For null character
            LPWSTR lpName = new WCHAR[cMaxSubKeyLen];
            DWORD cName;

            for (DWORD i = 0; i < cSubKeys; ++i) {
                cName = cMaxSubKeyLen;
                res = ::RegEnumKeyExW(handle, i, lpName, &cName, NULL, NULL, NULL, NULL);
                if (res == ERROR_SUCCESS)
                    keys.push_back(QString::fromWCharArray(lpName));
            }

            delete[] lpName;
        }

        return keys;
    }

    QString getRegValue(const HKEY handle, const QString &name = QString())
    {
        QString result;

        DWORD type = 0;
        DWORD cbData = 0;
        LPWSTR lpValueName = NULL;
        if (!name.isEmpty()) {
            lpValueName = new WCHAR[name.size() + 1];
            name.toWCharArray(lpValueName);
            lpValueName[name.size()] = 0;
        }

        // Discover the size of the value
        ::RegQueryValueExW(handle, lpValueName, NULL, &type, NULL, &cbData);
        DWORD cBuffer = (cbData / sizeof(WCHAR)) + 1;
        LPWSTR lpData = new WCHAR[cBuffer];
        LONG res = ::RegQueryValueExW(handle, lpValueName, NULL, &type, (LPBYTE)lpData, &cbData);
        if (lpValueName)
            delete[] lpValueName;

        if (res == ERROR_SUCCESS) {
            lpData[cBuffer - 1] = 0;
            result = QString::fromWCharArray(lpData);
        }
        delete[] lpData;

        return result;
    }

    QString miktexSearchReg(const REG_SEARCH_TYPE type)
    {
        HKEY hkRoot;
        if (type == USER)
            hkRoot = HKEY_CURRENT_USER;
        else
            hkRoot = HKEY_LOCAL_MACHINE;

        REGSAM samDesired = KEY_READ;
        if (type == SYSTEM_32BIT)
            samDesired |= KEY_WOW64_32KEY;
        else if (type == SYSTEM_64BIT)
            samDesired |= KEY_WOW64_64KEY;

        QString path, miktex_root;
        LONG res = 0;
        HKEY hkMiktex;
        res = ::RegOpenKeyExW(hkRoot, L"SOFTWARE\\MiKTeX.org\\MiKTeX", 0, samDesired, &hkMiktex);

        if (res == ERROR_SUCCESS) {
            QStringList versions = getRegSubkeys(hkMiktex);
            versions.sort();


            bool found = false;
            if  (!versions.empty()) {
                const QString core = versions.takeLast() + "\\Core";
                LPWSTR lpSubkeyCore = new WCHAR[core.size() + 1];
                core.toWCharArray(lpSubkeyCore);
                lpSubkeyCore[core.size()] = 0;
                HKEY hkMiktexCore;
                res = ::RegOpenKeyExW(hkMiktex, lpSubkeyCore, 0, samDesired, &hkMiktexCore);
                delete[] lpSubkeyCore;
                if (res == ERROR_SUCCESS) {
                    miktex_root= getRegValue(hkMiktexCore,"UserInstall");
                    ::RegCloseKey(hkMiktexCore);
                    path=miktex_root+"\\miktex\\bin\\x64";
                    if (!path.isEmpty() && QDir(path).exists("latex.exe")) {
                                            found = true;
                                            path = QDir(path).absolutePath();
                                        }
                    if (!found) {
                    path=miktex_root+"\\miktex\\bin";
                    if (!path.isEmpty() && QDir(path).exists("latex.exe")) {
                                            found = true;
                                            path = QDir(path).absolutePath();
                                        }
                         }
                }
            }
            if (!found)
                path = QString();

            ::RegCloseKey(hkMiktex);
        }

        return path;
    }

    QString findMiktexPath()
    {

        QString path = miktexSearchReg(USER);
        if (!path.isEmpty())
            return path;

        path = miktexSearchReg(SYSTEM_32BIT);
        if (!path.isEmpty())
            return path;

        path = miktexSearchReg(SYSTEM_64BIT);
        if (!path.isEmpty())
            return path;

        path=searchInWindowPath("miktex/bin");
        if (!path.isEmpty())
            return path;

        return QString("C:/Program Files/MiKTeX 2.9/miktex/bin/x64");
    }


    QString nativeAsymptoteSearchReg(const REG_SEARCH_TYPE type)
    {
        HKEY hkRoot;
        if (type == USER)
            hkRoot = HKEY_CURRENT_USER;
        else
            hkRoot = HKEY_LOCAL_MACHINE;

        REGSAM samDesired = KEY_READ;
        if (type == SYSTEM_32BIT)
            samDesired |= KEY_WOW64_32KEY;
        else if (type == SYSTEM_64BIT)
            samDesired |= KEY_WOW64_64KEY;

        QString path;
        LONG res = 0;
        HKEY hkAsymptote;
        res = ::RegOpenKeyExW(hkRoot, L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\Asymptote", 0, samDesired, &hkAsymptote);

        if (res == ERROR_SUCCESS) {
           bool found = false;
                    path= getRegValue(hkAsymptote,"Path");
                    if (!path.isEmpty() && QDir(path).exists("asy.exe"))
                        {
                        found = true;
                        path = QDir(path).absolutePath();
                        }
            if (!found)  path = QString();
            ::RegCloseKey(hkAsymptote);
        }

        return path;
    }
    QString findNativeAsymptotePath()
    {
        QString path = nativeAsymptoteSearchReg(USER);
        if (!path.isEmpty())
            return path;

        path = nativeAsymptoteSearchReg(SYSTEM_32BIT);
        if (!path.isEmpty())
            return path;

        path = nativeAsymptoteSearchReg(SYSTEM_64BIT);
        if (!path.isEmpty())
            return path;

        return QString();
    }

    QString findAsymptotePath()
    {
        QString path=findNativeAsymptotePath();
        if (!path.isEmpty())
            return path;

        path="C:/Program Files (x86)/Asymptote";
        if (QDir(path).exists("asy.exe"))
            return path;

        path="C:/Program Files/Asymptote";
        if (QDir(path).exists("asy.exe"))
            return path;

        path=searchInWindowPath("texlive");
        if (!path.isEmpty() && QDir(path).exists("asy.exe"))
            return path;

        path=searchInWindowPath("miktex/bin");
        if (!path.isEmpty() && QDir(path).exists("asy.exe"))
            return path;


        return QString("");
    }

    QString acrobatExeSearchReg(const REG_SEARCH_TYPE type)
    {
        HKEY hkRoot=HKEY_CLASSES_ROOT;


        REGSAM samDesired = KEY_READ;
        if (type == SYSTEM_32BIT)
            samDesired |= KEY_WOW64_32KEY;
        else if (type == SYSTEM_64BIT)
            samDesired |= KEY_WOW64_64KEY;

        QString path;
        LONG res = 0;
        HKEY hkAcrobat;
        res = ::RegOpenKeyExW(hkRoot, L"SOFTWARE\\Adobe\\Acrobat\\Exe", 0, samDesired, &hkAcrobat);

        if (res == ERROR_SUCCESS) {
           bool found = false;
                    path= getRegValue(hkAcrobat);
                    if (!path.isEmpty() )
                        {
                        path=path.remove("\"");
                        path=QFileInfo(path).absoluteFilePath();
                        if (QFileInfo(path).exists()) found=true;

                        }
            if (!found)  path = QString();
            ::RegCloseKey(hkAcrobat);
        }
        return path;
    }

    QString findAcrobatExePath()
    {
        QString path = acrobatExeSearchReg(SYSTEM_32BIT);
        if (!path.isEmpty())
            return path;

        path = acrobatExeSearchReg(SYSTEM_64BIT);
        if (!path.isEmpty())
             return path;


        return QString("C:/Program Files (x86)/Adobe/Acrobat Reader DC/Reader/AcroRd32.exe");
    }

    QString ghostscriptSearchReg(const REG_SEARCH_TYPE type)
    {
        HKEY hkRoot;
        if (type == USER)
            hkRoot = HKEY_CURRENT_USER;
        else
            hkRoot = HKEY_LOCAL_MACHINE;

        REGSAM samDesired = KEY_READ;
        if (type == SYSTEM_32BIT)
            samDesired |= KEY_WOW64_32KEY;
        else if (type == SYSTEM_64BIT)
            samDesired |= KEY_WOW64_64KEY;

        QString path, gs_root;
        LONG res = 0;
        HKEY hkGS;
        res = ::RegOpenKeyExW(hkRoot, L"SOFTWARE\\GPL Ghostscript", 0, samDesired, &hkGS);

        if (res == ERROR_SUCCESS) {
            QStringList versions = getRegSubkeys(hkGS);
            versions.sort();


            bool found = false;
            if  (!versions.empty()) {
                const QString core = versions.takeLast() ;
                LPWSTR lpSubkeyCore = new WCHAR[core.size() + 1];
                core.toWCharArray(lpSubkeyCore);
                lpSubkeyCore[core.size()] = 0;
                HKEY hkGSCore;
                res = ::RegOpenKeyExW(hkGS, lpSubkeyCore, 0, samDesired, &hkGSCore);
                delete[] lpSubkeyCore;
                if (res == ERROR_SUCCESS) {
                    gs_root= getRegValue(hkGSCore,"GS_LIB");
                    ::RegCloseKey(hkGSCore);
                    #if QT_VERSION < QT_VERSION_CHECK(5, 14, 0)
                    QStringList paths = gs_root.split(";", QString::SkipEmptyParts);
                    #else
                    QStringList paths = gs_root.split(";", Qt::SkipEmptyParts);
                    #endif
                    if (!paths.empty())
                    {
                    QString binpath=paths.takeFirst();
                    if (!binpath.isEmpty() && QDir(binpath).exists("gswin64c.exe")) {
                                            found = true;
                                            path = QDir(binpath).absolutePath()+"/gswin64c.exe";
                                        }
                    if (!found) {
                    if (!binpath.isEmpty() && QDir(binpath).exists("gswin32c.exe")) {
                                            found = true;
                                            path = QDir(binpath).absolutePath()+"/gswin32c.exe";
                                        }
                         }
                    }
                }
            }
            if (!found)
                path = QString();

            ::RegCloseKey(hkGS);
        }

        return path;
    }

    QString findGhostscriptPath()
    {

        QString path = ghostscriptSearchReg(USER);
        if (!path.isEmpty())
            return path;

        path = ghostscriptSearchReg(SYSTEM_32BIT);
        if (!path.isEmpty())
            return path;

        path = ghostscriptSearchReg(SYSTEM_64BIT);
        if (!path.isEmpty())
            return path;

        return QString("gswin64c.exe");
    }


#endif // Q_OS_WIN

MainWindow::MainWindow(QWidget *parent)
    :QMainWindow( parent)
{
fichier="";
ui.setupUi(this);

#if defined(Q_OS_MAC)
setWindowIcon(QIcon(":/images/logoabig.png"));
#else
setWindowIcon(QIcon(":/images/logosmall.png"));
#endif
QTextCodec *codec = QTextCodec::codecForName("UTF-8");
QString contenu;
QFile apropos(":/documents/presentation.txt");
apropos.open(QIODevice::ReadOnly);
contenu=codec->toUnicode(apropos.readAll());
apropos.close();
ui.textEdit->setOpenExternalLinks(true);
ui.textEdit->setHtml(contenu);

ui.textBrowser->setOpenExternalLinks(true);
ui.textBrowser->setHtml(QString::fromUtf8("<b>PDfAdd ")+QLatin1String(VERSION_STR)+QString::fromUtf8("</b><br>Assistants de création de  graphiques pour documents mathématiques composés avec pdflatex et asymptote.<br><b><i>Copyright (c) 2012-2021 par Pascal Brachet</i></b><br>Ce programme libre est distribué selon les termes de la licence GPL.<br>PdfAdd contient du code venant des programmes : Pdfium (BSD), QtPDF (The Qt Company Ltd - GPL), Eukleides (Copyright (c) Christian Obrecht), QtCreator (Copyright (C) Nokia - Licence: GPL) et du code LaTeX/Asymptote a été adapté des travaux de D.Léger, C.Grospellier et P.Ivaldi."));

QFile file(":/documents/styleapplication.qss");
file.open(QFile::ReadOnly);
QString styleSheet = QLatin1String(file.readAll());
qApp->setStyleSheet(styleSheet);

connect( ui.pushButtonPdfLatex, SIGNAL(clicked()), this, SLOT(chercherPdfLatex()));
connect( ui.pushButtonAsy, SIGNAL(clicked()), this, SLOT(chercherAsy()));
connect( ui.pushButtonLecteurpdf, SIGNAL(clicked()), this, SLOT(chercherLecteurpdf()));
connect( ui.pushButtonCouleur1, SIGNAL(clicked()), this, SLOT(configurerCouleur1()));
connect( ui.pushButtonCouleur2, SIGNAL(clicked()), this, SLOT(configurerCouleur2()));
connect( ui.pushButtonCouleur3, SIGNAL(clicked()), this, SLOT(configurerCouleur3()));
connect( ui.pushButtonRAZ, SIGNAL(clicked()), this, SLOT(reinitConfig()));

QFont commandFont=qApp->font();
QFont titleFont = commandFont;
titleFont.setBold(true);

QColor titleBg("#447BCD");
QColor titleFg("#ffffff");

ui.listWidget->viewport()->setAttribute(Qt::WA_Hover, true);
ui.listWidget->setItemDelegate(new ListDelegate(ui.listWidget));

itemAnalyse = new QListWidgetItem(ui.listWidget);
itemAnalyse->setText(QString::fromUtf8("ANALYSE"));
itemAnalyse->setBackground(titleBg);
itemAnalyse->setForeground(titleFg);
itemAnalyse->setFont(titleFont);
itemAnalyse->setTextAlignment(Qt::AlignHCenter);

itemCourbe = new QListWidgetItem(ui.listWidget);
itemCourbe->setText(QString::fromUtf8("Courbes 2D"));
itemCourbe->setFont(commandFont);
//itemCourbe->setSizeHint(QSize(225,25));

itemTableau = new QListWidgetItem(ui.listWidget);
itemTableau->setText(QString::fromUtf8("Tableau de variations"));
itemTableau->setFont(commandFont);
//itemTableau->setSizeHint(QSize(225,25));

itemSuite = new QListWidgetItem(ui.listWidget);
itemSuite->setText(QString::fromUtf8("Suite récurrente"));
itemSuite->setFont(commandFont);
//itemSuite->setSizeHint(QSize(225,25));

itemStat = new QListWidgetItem(ui.listWidget);
itemStat->setText(QString::fromUtf8("STATISTIQUE"));
itemStat->setBackground(titleBg);
itemStat->setForeground(titleFg);
itemStat->setFont(titleFont);
itemStat->setTextAlignment(Qt::AlignHCenter);

itemBoite = new QListWidgetItem(ui.listWidget);
itemBoite->setText(QString::fromUtf8("Diagramme en boîtes"));
itemBoite->setFont(commandFont);
//itemBoite->setSizeHint(QSize(225,25));

itemHisto = new QListWidgetItem(ui.listWidget);
itemHisto->setText(QString::fromUtf8("Histogramme"));
itemHisto->setFont(commandFont);
//itemHisto->setSizeHint(QSize(225,25));

itemProba = new QListWidgetItem(ui.listWidget);
itemProba->setText(QString::fromUtf8("PROBABILITÉS"));
itemProba->setBackground(titleBg);
itemProba->setForeground(titleFg);
itemProba->setFont(titleFont);
itemProba->setTextAlignment(Qt::AlignHCenter);

itemArbre = new QListWidgetItem(ui.listWidget);
itemArbre->setText(QString::fromUtf8("Arbres pondérés"));
itemArbre->setFont(commandFont);
//itemArbre->setSizeHint(QSize(225,25));

itemBinomiale = new QListWidgetItem(ui.listWidget);
itemBinomiale->setText(QString::fromUtf8("Loi binomiale"));
itemBinomiale->setFont(commandFont);
//itemSurface->setSizeHint(QSize(225,25));

itemAutre = new QListWidgetItem(ui.listWidget);
itemAutre->setText(QString::fromUtf8("DIVERS"));
itemAutre->setBackground(titleBg);
itemAutre->setForeground(titleFg);
itemAutre->setFont(titleFont);
itemAutre->setTextAlignment(Qt::AlignHCenter);

itemGraphe = new QListWidgetItem(ui.listWidget);
itemGraphe->setText(QString::fromUtf8("Graphes"));
itemGraphe->setFont(commandFont);
//itemGraphe->setSizeHint(QSize(225,25));

itemCercle = new QListWidgetItem(ui.listWidget);
itemCercle->setText(QString::fromUtf8("Cercle trigonométrique"));
itemCercle->setFont(commandFont);
//itemGraphe->setSizeHint(QSize(225,25));

itemGeometrie = new QListWidgetItem(ui.listWidget);
itemGeometrie->setText(QString::fromUtf8("GÉOMETRIE"));
itemGeometrie->setBackground(titleBg);
itemGeometrie->setForeground(titleFg);
itemGeometrie->setFont(titleFont);
itemGeometrie->setTextAlignment(Qt::AlignHCenter);

itemGeo = new QListWidgetItem(ui.listWidget);
itemGeo->setText(QString::fromUtf8("Figures géométriques"));
itemGeo->setFont(commandFont);
//itemGeo->setSizeHint(QSize(225,25));

itemEspace = new QListWidgetItem(ui.listWidget);
itemEspace->setText(QString::fromUtf8("Repère dans l'espace"));
itemEspace->setFont(commandFont);
//itemEspace->setSizeHint(QSize(225,25));

LireConfig();
extra_path="";
#if defined(Q_OS_MAC)
extra_path=":/Library/TeX/texbin:/Library/TeX/Distributions/.DefaultTeX/Contents/Programs/texbin:/usr/bin:/usr/sbin:/sbin:/usr/local/bin:/usr/texbin:/sw/bin:";
#endif
#if defined(Q_OS_WIN32)
if (!searchInWindowPath("texlive").isEmpty()) extra_path=extra_path+";"+searchInWindowPath("texlive");
else extra_path=extra_path+";"+findMiktexPath();
#endif

ui.lineEditPdfLatex->setText(pdflatexCd);
ui.lineEditAsy->setText(asyCd);
ui.lineEditLecteurpdf->setText(lecteurpdfCd);
ui.pushButtonCouleur1->setPalette(QPalette(couleur1));
ui.pushButtonCouleur1->setAutoFillBackground(true);
ui.pushButtonCouleur2->setPalette(QPalette(couleur2));
ui.pushButtonCouleur2->setAutoFillBackground(true);
ui.pushButtonCouleur3->setPalette(QPalette(couleur3));
ui.pushButtonCouleur3->setAutoFillBackground(true);
ui.textEditAsyPre->setPlainText(preamAsy);

QAction *Act;

aslatMenu = menuBar()->addMenu(QString::fromUtf8("PdfAdd"));
Act = new QAction(QIcon(":/images/configurer.png"), QString::fromUtf8("Configurer"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Configurer()));
aslatMenu->addAction(Act);
aslatMenu->addSeparator();
Act = new QAction(QString::fromUtf8("Ouvrir un exemple"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(ChargerExemple()));
aslatMenu->addAction(Act);
aslatMenu->addSeparator();
#if defined(Q_OS_UNIX) && !defined(Q_OS_MAC)
Act = new QAction(QString::fromUtf8("Changer Police Interface"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(SetInterfaceFont()));
aslatMenu->addAction(Act);
aslatMenu->addSeparator();
#endif
Act = new QAction(QIcon(":/images/quitter16.png"), QString::fromUtf8("Quitter"), this);
Act->setShortcut(Qt::CTRL | Qt::Key_Q);
connect(Act, SIGNAL(triggered()), this, SLOT(Quitter()));
aslatMenu->addAction(Act);

assistantMenu = menuBar()->addMenu(QString::fromUtf8("Assistants"));
Act = new QAction(QString::fromUtf8("Courbes 2D"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Courbes()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Tableaux de variations"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Tabvar()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Arbres pondérés"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Arbre()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Figures géométriques"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Geometrie()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Graphes"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Graphe()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Diagramme en boîtes"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Boites()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Suite récurrente"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Suite()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Histogramme"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Histogramme()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Loi binomiale"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Binomiale()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Repère dans l'espace"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Espace()));
assistantMenu->addAction(Act);
Act = new QAction(QString::fromUtf8("Cercle trigonométrique"), this);
connect(Act, SIGNAL(triggered()), this, SLOT(Cercle()));
assistantMenu->addAction(Act);

connect(ui.listWidget, SIGNAL(itemClicked ( QListWidgetItem*)), this, SLOT(LancerAssistant(QListWidgetItem*)));
ui.tabWidget->setCurrentIndex(0);

highlighter = new Highlighter(ui.textEditAsyPre->document());

QFont efont;
QFontDatabase fdb;
QStringList xf = fdb.families();
QString deft;
if (xf.contains("Liberation Mono",Qt::CaseInsensitive)) deft="Liberation Mono";
else if (xf.contains("DejaVu Sans Mono",Qt::CaseInsensitive)) deft="DejaVu Sans Mono";
#if defined(Q_OS_MAC)
else if (xf.contains("PT Mono",Qt::CaseInsensitive)) deft="PT Mono";
else if (xf.contains("Andale Mono",Qt::CaseInsensitive)) deft="Andale Mono";
else if (xf.contains("Monaco",Qt::CaseInsensitive)) deft="Monaco";
#endif
#if defined(Q_OS_WIN32)
else if (xf.contains("Courier New",Qt::CaseInsensitive)) deft="Courier New";
#endif
else deft=qApp->font().family();  
efont.setFamily(deft);
efont.setPointSize(qApp->font().pointSize());
ui.textEditAsyPre->setFont(efont);
ui.textEditAsyPre->setWordWrapMode(QTextOption::NoWrap);
}

MainWindow::~MainWindow(){
}

void MainWindow::closeEvent(QCloseEvent *e)
{
SauverConfig();
e->accept();
}

void MainWindow::Quitter()
{
SauverConfig();
qApp->quit();
}

void MainWindow::Ouvrir(QString f )
{
QDomDocument doc;
if (f.isEmpty()) return;
QFileInfo fi(f);
if (!fi.exists()) return;
QFile fich(f);
if (!fich.open(QIODevice::ReadOnly)) 
	{
	QMessageBox::warning( this,QString::fromUtf8("Erreur"),QString::fromUtf8("Impossible de lire le fichier"));
	return;
	}
if (!doc.setContent(&fich)) 
	{
	QMessageBox::warning( this,QString::fromUtf8("Erreur"),QString::fromUtf8("Impossible de lire le fichier"));
	fich.close();
	return;
	}
QDomElement root = doc.documentElement();
if (root.tagName() == "Courbes")
	{
	fich.close();
	fichier=f;
	Courbes();
	}
else if (root.tagName() == "Tableau")
	{
	fich.close();
	fichier=f;
	Tabvar();
	}
else if (root.tagName() == "Arbre")
	{
	fich.close();
	fichier=f;
	Arbre();
	}
else if (root.tagName() == "Geometrie")
	{
	fich.close();
	fichier=f;
	Geometrie();
	}
else if (root.tagName() == "Graphe")
	{
	fich.close();
	fichier=f;
	Graphe();
	}
else if (root.tagName() == "Boites")
	{
	fich.close();
	fichier=f;
	Boites();
	}
else if (root.tagName() == "Histogramme")
	{
	fich.close();
	fichier=f;
	Histogramme();
	}
else if (root.tagName() == "Suite")
	{
	fich.close();
	fichier=f;
	Suite();
	}
else if (root.tagName() == "Binomiale")
	{
	fich.close();
	fichier=f;
	Binomiale();
	}
else if (root.tagName() == "Espace")
	{
	fich.close();
	fichier=f;
	Espace();
	}
else if (root.tagName() == "Cercle")
	{
	fich.close();
	fichier=f;
	Cercle();
	}
else
	{
	fich.close();
	QMessageBox::warning( this,QString::fromUtf8("Erreur"),QString::fromUtf8("Le fichier ne correspond à aucun des assistants"));
	}

}

//*****************************************
void MainWindow::Courbes()
{
ActualiserConfig();
CourbesDialog *courbesDialog = new CourbesDialog(this);
courbesDialog->extraPath=extra_path;
courbesDialog->asyCommande=asyCd;
courbesDialog->pdflatexCommande=pdflatexCd;
courbesDialog->lecteurpdfCommande=lecteurpdfCd;
courbesDialog->dernierRepertoire=dernierRp;
courbesDialog->couleurSupp1=couleur1;
courbesDialog->couleurSupp2=couleur2;
courbesDialog->couleurSupp3=couleur3;
courbesDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) courbesDialog->Ouvrir(fichier);
if (courbesDialog->exec()) dernierRp=courbesDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Tabvar()
{
ActualiserConfig();
TabvarDialog *tabvarDialog = new TabvarDialog(this);
tabvarDialog->extraPath=extra_path;
tabvarDialog->asyCommande=asyCd;
tabvarDialog->pdflatexCommande=pdflatexCd;
tabvarDialog->lecteurpdfCommande=lecteurpdfCd;
tabvarDialog->dernierRepertoire=dernierRp;
if (!fichier.isEmpty()) tabvarDialog->Ouvrir(fichier);
if (tabvarDialog->exec()) dernierRp=tabvarDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Arbre()
{
ActualiserConfig();
ArbreDialog *arbreDialog = new ArbreDialog(this);
arbreDialog->extraPath=extra_path;
arbreDialog->asyCommande=asyCd;
arbreDialog->pdflatexCommande=pdflatexCd;
arbreDialog->lecteurpdfCommande=lecteurpdfCd;
arbreDialog->dernierRepertoire=dernierRp;
arbreDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) arbreDialog->Ouvrir(fichier);
if (arbreDialog->exec()) dernierRp=arbreDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Geometrie()
{
ActualiserConfig();
GeometrieDialog *geoDialog = new GeometrieDialog(this);
geoDialog->extraPath=extra_path;
geoDialog->asyCommande=asyCd;
geoDialog->pdflatexCommande=pdflatexCd;
geoDialog->lecteurpdfCommande=lecteurpdfCd;
geoDialog->dernierRepertoire=dernierRp;
geoDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) geoDialog->Ouvrir(fichier);
if (geoDialog->exec()) dernierRp=geoDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Graphe()
{
ActualiserConfig();
GrapheDialog *grapheDialog = new GrapheDialog(this);
grapheDialog->extraPath=extra_path;
grapheDialog->asyCommande=asyCd;
grapheDialog->pdflatexCommande=pdflatexCd;
grapheDialog->lecteurpdfCommande=lecteurpdfCd;
grapheDialog->dernierRepertoire=dernierRp;
grapheDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) grapheDialog->Ouvrir(fichier);
if (grapheDialog->exec()) dernierRp=grapheDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Boites()
{
ActualiserConfig();
BoitesDialog *boitesDialog = new BoitesDialog(this);
boitesDialog->extraPath=extra_path;
boitesDialog->asyCommande=asyCd;
boitesDialog->pdflatexCommande=pdflatexCd;
boitesDialog->lecteurpdfCommande=lecteurpdfCd;
boitesDialog->dernierRepertoire=dernierRp;
boitesDialog->couleurSupp1=couleur1;
boitesDialog->couleurSupp2=couleur2;
boitesDialog->couleurSupp3=couleur3;
boitesDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) boitesDialog->Ouvrir(fichier);
if (boitesDialog->exec()) dernierRp=boitesDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Suite()
{
ActualiserConfig();
SuitesDialog *suitesDialog = new SuitesDialog(this);
suitesDialog->extraPath=extra_path;
suitesDialog->asyCommande=asyCd;
suitesDialog->pdflatexCommande=pdflatexCd;
suitesDialog->lecteurpdfCommande=lecteurpdfCd;
suitesDialog->dernierRepertoire=dernierRp;
suitesDialog->couleurSupp1=couleur1;
suitesDialog->couleurSupp2=couleur2;
suitesDialog->couleurSupp3=couleur3;
suitesDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) suitesDialog->Ouvrir(fichier);
if (suitesDialog->exec()) dernierRp=suitesDialog->dernierRepertoire;
fichier="";

}

void MainWindow::Histogramme()
{
ActualiserConfig();
HistoDialog *histoDialog = new HistoDialog(this);
histoDialog->extraPath=extra_path;
histoDialog->asyCommande=asyCd;
histoDialog->pdflatexCommande=pdflatexCd;
histoDialog->lecteurpdfCommande=lecteurpdfCd;
histoDialog->dernierRepertoire=dernierRp;
histoDialog->couleurSupp1=couleur1;
histoDialog->couleurSupp2=couleur2;
histoDialog->couleurSupp3=couleur3;
histoDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) histoDialog->Ouvrir(fichier);
if (histoDialog->exec()) dernierRp=histoDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Binomiale()
{
ActualiserConfig();
BinomDialog *binomDialog = new BinomDialog(this);
binomDialog->extraPath=extra_path;
binomDialog->asyCommande=asyCd;
binomDialog->pdflatexCommande=pdflatexCd;
binomDialog->lecteurpdfCommande=lecteurpdfCd;
binomDialog->dernierRepertoire=dernierRp;
binomDialog->couleurSupp1=couleur1;
binomDialog->couleurSupp2=couleur2;
binomDialog->couleurSupp3=couleur3;
binomDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) binomDialog->Ouvrir(fichier);
if (binomDialog->exec()) dernierRp=binomDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Espace()
{
ActualiserConfig();
EspaceDialog *espaceDialog = new EspaceDialog(this);
espaceDialog->extraPath=extra_path;
espaceDialog->asyCommande=asyCd;
espaceDialog->pdflatexCommande=pdflatexCd;
espaceDialog->lecteurpdfCommande=lecteurpdfCd;
espaceDialog->dernierRepertoire=dernierRp;
espaceDialog->couleurSupp1=couleur1;
espaceDialog->couleurSupp2=couleur2;
espaceDialog->couleurSupp3=couleur3;
espaceDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) espaceDialog->Ouvrir(fichier);
if (espaceDialog->exec()) dernierRp=espaceDialog->dernierRepertoire;
fichier="";
}

void MainWindow::Cercle()
{
ActualiserConfig();
TrigoDialog *trigoDialog = new TrigoDialog(this);
trigoDialog->extraPath=extra_path;
trigoDialog->asyCommande=asyCd;
trigoDialog->pdflatexCommande=pdflatexCd;
trigoDialog->lecteurpdfCommande=lecteurpdfCd;
trigoDialog->dernierRepertoire=dernierRp;
trigoDialog->couleurSupp1=couleur1;
trigoDialog->couleurSupp2=couleur2;
trigoDialog->couleurSupp3=couleur3;
trigoDialog->preambuleAsy=preamAsy;
if (!fichier.isEmpty()) trigoDialog->Ouvrir(fichier);
if (trigoDialog->exec()) dernierRp=trigoDialog->dernierRepertoire;
fichier="";
}

//*************************
void MainWindow::LireConfig()
{
#ifdef USB_VERSION
QSettings *config=new QSettings(QCoreApplication::applicationDirPath()+"/pdfadd.ini",QSettings::IniFormat); //for USB-stick version :
#else
QSettings *config=new QSettings(QSettings::IniFormat,QSettings::UserScope,"xm1","pdfadd");
#endif
/*if (!config->contains("IniMode")) 
	{
	delete config;
	config=new QSettings("xm1","pstplus");
	}*/

config->beginGroup( "pdfadd" );
#if defined(Q_OS_MAC)
pdflatexCd=config->value("commande pdflatex","\"pdflatex\"").toString();
asyCd=config->value("commande asy","\"asy\"").toString();
lecteurpdfCd=config->value("commande lecteurpdf","open").toString();
qApp->setStyle(QLatin1String("fusion"));
#endif
#if defined(Q_OS_UNIX) && !defined(Q_OS_MAC)
int desktop_env=1; // 1 : gnome ; 2: kde ; 3 : kde4 ; 
QStringList styles = QStyleFactory::keys();
QString kdesession= ::getenv("KDE_FULL_SESSION");
QString kdeversion= ::getenv("KDE_SESSION_VERSION");
if (!kdesession.isEmpty()) desktop_env=2;
if (!kdeversion.isEmpty()) desktop_env=3;

switch (desktop_env)
	{
	case 1:
		{
		lecteurpdfCd=config->value("commande lecteurpdf","evince").toString();
		}break;
	case 2:
		{
		lecteurpdfCd=config->value("commande lecteurpdf","kpdf").toString();
		}break;
	case 3:
		{
		lecteurpdfCd=config->value("commande lecteurpdf","okular").toString();
		}break;
	}

pdflatexCd=config->value("commande pdflatex","pdflatex").toString();
asyCd=config->value("commande asy","asy").toString();
x11style=config->value( "X11/Style","Plastique").toString();
QFontDatabase fdb;
QStringList xf = fdb.families();
QString deft;
if (xf.contains("DejaVu Sans",Qt::CaseInsensitive)) deft="DejaVu Sans";
else if (xf.contains("DejaVu Sans LGC",Qt::CaseInsensitive)) deft="DejaVu Sans LGC";
else if (xf.contains("Bitstream Vera Sans",Qt::CaseInsensitive)) deft="Bitstream Vera Sans";
else if (xf.contains("Luxi Sans",Qt::CaseInsensitive)) deft="Luxi Sans";
else deft=qApp->font().family();
x11fontfamily=config->value("X11/Font Family",deft).toString();
x11fontsize=config->value( "X11/Font Size","10").toInt();

if(desktop_env == 1)
{
if ((styles.contains("GTK+")) || (styles.contains("gtk2"))) qApp->setStyle(QLatin1String("gtkstyle"));
else if (styles.contains("Breeze")) qApp->setStyle(QLatin1String("breeze"));
else qApp->setStyle(QLatin1String("fusion"));    
}
else if (styles.contains("Breeze")) qApp->setStyle(QLatin1String("breeze"));
else qApp->setStyle(QLatin1String("fusion"));

QFont x11Font (x11fontfamily,x11fontsize);
QApplication::setFont(x11Font);
#ifdef STATIC_VERSION
QPalette pal = QApplication::palette();

pal.setColor( QPalette::Active, QPalette::Highlight, QColor("#4490d8") );
pal.setColor( QPalette::Inactive, QPalette::Highlight, QColor("#4490d8") );
pal.setColor( QPalette::Disabled, QPalette::Highlight, QColor("#4490d8") );


//pal.setColor( QPalette::Active, QPalette::Highlight, QColor("#5689be") );
//pal.setColor( QPalette::Inactive, QPalette::Highlight, QColor("#5689be") );
//pal.setColor( QPalette::Disabled, QPalette::Highlight, QColor("#5689be") );

pal.setColor( QPalette::Active, QPalette::HighlightedText, QColor("#ffffff") );
pal.setColor( QPalette::Inactive, QPalette::HighlightedText, QColor("#ffffff") );
pal.setColor( QPalette::Disabled, QPalette::HighlightedText, QColor("#ffffff") );

pal.setColor( QPalette::Active, QPalette::Base, QColor("#ffffff") );
pal.setColor( QPalette::Inactive, QPalette::Base, QColor("#ffffff") );
pal.setColor( QPalette::Disabled, QPalette::Base, QColor("#ffffff") );

pal.setColor( QPalette::Active, QPalette::WindowText, QColor("#000000") );
pal.setColor( QPalette::Inactive, QPalette::WindowText, QColor("#000000") );
pal.setColor( QPalette::Disabled, QPalette::WindowText, QColor("#000000") );

pal.setColor( QPalette::Active, QPalette::ButtonText, QColor("#000000") );
pal.setColor( QPalette::Inactive, QPalette::ButtonText, QColor("#000000") );
pal.setColor( QPalette::Disabled, QPalette::ButtonText, QColor("#000000") );

if (desktop_env ==3)
	{
	pal.setColor( QPalette::Active, QPalette::Window, QColor("#eae9e9") );
	pal.setColor( QPalette::Inactive, QPalette::Window, QColor("#eae9e9") );
	pal.setColor( QPalette::Disabled, QPalette::Window, QColor("#eae9e9") );

	pal.setColor( QPalette::Active, QPalette::Button, QColor("#eae9e9") );
	pal.setColor( QPalette::Inactive, QPalette::Button, QColor("#eae9e9") );
	pal.setColor( QPalette::Disabled, QPalette::Button, QColor("#eae9e9") );
	}
else
	{
/*
	pal.setColor( QPalette::Active, QPalette::Window, QColor("#f1f1f1") );
	pal.setColor( QPalette::Inactive, QPalette::Window, QColor("#f1f1f1") );
	pal.setColor( QPalette::Disabled, QPalette::Window, QColor("#f1f1f1") );

	pal.setColor( QPalette::Active, QPalette::Button, QColor("#f1f1f1") );
	pal.setColor( QPalette::Inactive, QPalette::Button, QColor("#f1f1f1") );
	pal.setColor( QPalette::Disabled, QPalette::Button, QColor("#f1f1f1") );
*/
	pal.setColor( QPalette::Active, QPalette::Window, QColor("#fbf8f1") );
	pal.setColor( QPalette::Inactive, QPalette::Window, QColor("#fbf8f1") );
	pal.setColor( QPalette::Disabled, QPalette::Window, QColor("#fbf8f1") );

	pal.setColor( QPalette::Active, QPalette::Button, QColor("#fbf8f1") );
	pal.setColor( QPalette::Inactive, QPalette::Button, QColor("#fbf8f1") );
	pal.setColor( QPalette::Disabled, QPalette::Button, QColor("#fbf8f1") );
	}

QApplication::setPalette(pal);
#endif
#endif
#if defined(Q_OS_WIN32)
pdflatexCd=config->value("commande pdflatex","pdflatex").toString();

if (!findAsymptotePath().isEmpty()) asyCd=config->value("commande asy","\""+findAsymptotePath()+"/asy.exe\"").toString();
else asyCd=config->value("commande asy","asy.exe").toString();
lecteurpdfCd=config->value("commande lecteurpdf","\""+findAcrobatExePath()+"\"").toString();
qApp->setStyle(QLatin1String("fusion"));
#endif
dernierRp=config->value("dernier repertoire",QDir::homePath()).toString();
preamAsy=config->value("preambule asy","usepackage(\"amsmath,amssymb\");\nusepackage(\"inputenc\",\"utf8\");\nusepackage(\"icomma\");").toString();
#if (QT_VERSION >= QT_VERSION_CHECK(5, 14, 0))
   QRect screen = QWidget::screen()->geometry();
 #else
  QRect screen = (window() && window()->windowHandle() ? window()->windowHandle()->screen()->geometry() : QGuiApplication::primaryScreen()->geometry());
#endif
int x= config->value( "Geometries/MainwindowX",10).toInt();
int y= config->value( "Geometries/MainwindowY",10).toInt() ;
move(x,y);
couleur1=config->value("couleur utilisateur1",QColor(167,167,0)).value<QColor>();
couleur2=config->value("couleur utilisateur2",QColor(0,81,121)).value<QColor>();
couleur3=config->value("couleur utilisateur3",QColor(24,165,5)).value<QColor>();
config->endGroup();
}

void MainWindow::SauverConfig()
{
#ifdef USB_VERSION
QSettings config(QCoreApplication::applicationDirPath()+"/pdfadd.ini",QSettings::IniFormat); //for USB-stick version 
#else
QSettings config(QSettings::IniFormat,QSettings::UserScope,"xm1","pdfadd");
#endif
ActualiserConfig();

config.setValue( "IniMode",true);
config.beginGroup( "pdfadd" );
config.setValue("commande pdflatex",pdflatexCd);
config.setValue("commande asy",asyCd);
config.setValue("commande lecteurpdf",lecteurpdfCd);
config.setValue("dernier repertoire",dernierRp);
config.setValue("preambule asy",preamAsy);
config.setValue("Geometries/MainwindowX", x() );
config.setValue("Geometries/MainwindowY", y() );
#if defined(Q_OS_UNIX) && !defined(Q_OS_MAC)
config.setValue( "X11/Style",x11style);
config.setValue("X11/Font Family",x11fontfamily);
config.setValue( "X11/Font Size",x11fontsize);
#endif
config.setValue("couleur utilisateur1",couleur1);
config.setValue("couleur utilisateur2",couleur2);
config.setValue("couleur utilisateur3",couleur3);
config.endGroup();
}

void MainWindow::Configurer()
{
ui.tabWidget->setCurrentIndex(1);
}

void MainWindow::reinitConfig()
{
#if defined(Q_OS_MAC)
pdflatexCd="\"pdflatex\"";
asyCd="\"asy\"";
lecteurpdfCd="open";
#endif
#if defined(Q_OS_UNIX) && !defined(Q_OS_MAC)
int desktop_env=1; // 1 : gnome ; 2: kde ; 3 : kde4 ;
QStringList styles = QStyleFactory::keys();
QString kdesession= ::getenv("KDE_FULL_SESSION");
QString kdeversion= ::getenv("KDE_SESSION_VERSION");
if (!kdesession.isEmpty()) desktop_env=2;
if (!kdeversion.isEmpty()) desktop_env=3;

switch (desktop_env)
	{
	case 1:
		{
		lecteurpdfCd="evince";
		}break;
	case 2:
		{
		lecteurpdfCd="kpdf";
		}break;
	case 3:
		{
		lecteurpdfCd="okular";
		}break;
	}

pdflatexCd="pdflatex";
asyCd="asy";
#endif
#if defined(Q_OS_WIN32)
pdflatexCd="pdflatex";
if (!findAsymptotePath().isEmpty()) asyCd="\""+findAsymptotePath()+"/asy.exe\"";
else asyCd="asy.exe";
lecteurpdfCd="\""+findAcrobatExePath()+"\"";
#endif
preamAsy="usepackage(\"amsmath,amssymb\");\nusepackage(\"inputenc\",\"utf8\");\nusepackage(\"icomma\");";
couleur1=QColor(167,167,0);
couleur2=QColor(0,81,121);
couleur3=QColor(24,165,5);

ui.lineEditPdfLatex->setText(pdflatexCd);
ui.lineEditAsy->setText(asyCd);
ui.lineEditLecteurpdf->setText(lecteurpdfCd);
ui.pushButtonCouleur1->setPalette(QPalette(couleur1));
ui.pushButtonCouleur1->setAutoFillBackground(true);
ui.pushButtonCouleur2->setPalette(QPalette(couleur2));
ui.pushButtonCouleur2->setAutoFillBackground(true);
ui.pushButtonCouleur3->setPalette(QPalette(couleur3));
ui.pushButtonCouleur3->setAutoFillBackground(true);
ui.textEditAsyPre->setPlainText(preamAsy);
}

void MainWindow::ActualiserConfig()
{
pdflatexCd=ui.lineEditPdfLatex->text();
asyCd=ui.lineEditAsy->text();
lecteurpdfCd=ui.lineEditLecteurpdf->text();
preamAsy=ui.textEditAsyPre->toPlainText();
couleur1=ui.pushButtonCouleur1->palette().window().color();
couleur2=ui.pushButtonCouleur2->palette().window().color();
couleur3=ui.pushButtonCouleur3->palette().window().color();
}
//**************************
void MainWindow::ChargerExemple()
{
QString exempleRepertoire="";
#if defined(Q_OS_UNIX) && !defined(Q_OS_MAC)
#ifdef USB_VERSION
exempleRepertoire=QCoreApplication::applicationDirPath();
#else
exempleRepertoire=PREFIX"/share/pdfadd";
#endif
#endif
#if defined(Q_OS_MAC)
exempleRepertoire=QCoreApplication::applicationDirPath()+"/../Resources";
//exempleRepertoire="/Applications/pstplus.app/Contents/Resources";
#endif
#if defined(Q_OS_WIN32)
exempleRepertoire=QCoreApplication::applicationDirPath();
#endif
QString f="";
f = QFileDialog::getOpenFileName(this,QString::fromUtf8("Ouvrir Fichier"),exempleRepertoire,"Fichier PdfAdd (*.pag)");
Ouvrir(f);
}
//*************************
void MainWindow::SetInterfaceFont()
{
#if defined(Q_OS_UNIX) && !defined(Q_OS_MAC)
X11FontDialog *xfdlg = new X11FontDialog(this);
int ft=xfdlg->ui.comboBoxFont->findText (x11fontfamily , Qt::MatchExactly);
xfdlg->ui.comboBoxFont->setCurrentIndex(ft);
xfdlg->ui.spinBoxSize->setValue(x11fontsize);
if (xfdlg->exec())
	{
	x11fontfamily=xfdlg->ui.comboBoxFont->currentText();
	x11fontsize=xfdlg->ui.spinBoxSize->value();
	QFont x11Font (x11fontfamily,x11fontsize);
	QApplication::setFont(x11Font);
	}
#endif
}
//*************************
void MainWindow::chercherPdfLatex()
{
QString location=QFileDialog::getOpenFileName(this,QString::fromUtf8("Chercher programme"),QDir::rootPath(),"Programme (*)",0,QFileDialog::DontResolveSymlinks);
if ( !location.isEmpty() ) 
	{
	location.replace(QString("\\"),QString("/"));
	location="\""+location+"\"";
	ui.lineEditPdfLatex->setText( location );
	}
}

void MainWindow::chercherAsy()
{
QString location=QFileDialog::getOpenFileName(this,QString::fromUtf8("Chercher programme"),QDir::rootPath(),"Programme (*)",0,QFileDialog::DontResolveSymlinks);
if ( !location.isEmpty() ) 
	{
	location.replace(QString("\\"),QString("/"));
	location="\""+location+"\"";
	ui.lineEditAsy->setText( location );
	}
}

void MainWindow::chercherLecteurpdf()
{
QString location=QFileDialog::getOpenFileName(this,QString::fromUtf8("Chercher programme"),QDir::rootPath(),"Programme (*)",0,QFileDialog::DontResolveSymlinks);
if ( !location.isEmpty() ) 
	{
	location.replace(QString("\\"),QString("/"));
	location="\""+location+"\"";
	ui.lineEditLecteurpdf->setText( location );
	}
}



void MainWindow::configurerCouleur1()
{
QColor color = QColorDialog::getColor(ui.pushButtonCouleur1->palette().window().color(), this);
if (color.isValid()) 
	{
	ui.pushButtonCouleur1->setPalette(QPalette(color));
	ui.pushButtonCouleur1->setAutoFillBackground(true);
	}
}

void MainWindow::configurerCouleur2()
{
QColor color = QColorDialog::getColor(ui.pushButtonCouleur2->palette().window().color(), this);
if (color.isValid()) 
	{
	ui.pushButtonCouleur2->setPalette(QPalette(color));
	ui.pushButtonCouleur2->setAutoFillBackground(true);
	}
}

void MainWindow::configurerCouleur3()
{
QColor color = QColorDialog::getColor(ui.pushButtonCouleur3->palette().window().color(), this);
if (color.isValid()) 
	{
	ui.pushButtonCouleur3->setPalette(QPalette(color));
	ui.pushButtonCouleur3->setAutoFillBackground(true);
	}
}

void MainWindow::LancerAssistant(QListWidgetItem* item)
{
if (item==itemCourbe) Courbes();
else if (item==itemSuite) Suite();
else if (item==itemTableau) Tabvar(); 
else if (item==itemBinomiale) Binomiale(); 
else if (item==itemGeo) Geometrie();
else if (item==itemEspace) Espace(); 
else if (item==itemBoite) Boites(); 
else if (item==itemHisto) Histogramme(); 
else if (item==itemArbre) Arbre();
else if (item==itemGraphe) Graphe();
else if (item==itemCercle) Cercle(); 
}


